﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CG2.Mathematics;
using System.Drawing;
using CG2.Modeling;
using CG2.Shading;
using CG2.Lighting;

namespace CG2.Rendering
{
	public class Camera
	{

        public struct HitPoint
        {
            public Vector4 Position;
            public Vector4 Color;
            public Vector4 Normal;
            public Double ShadowCount;
        }

		#region Properties

		public Vector4 Position;
		public Vector4 Target;
		public Vector4 Up = new Vector4(0, 0, 1);
		public Double FovY = 45;
        public Double zFar;
        public Double zNear;
		public Vector4 U, V, W; // camera to world space
		public Vector4 BgColor = new Vector4(0, 0, 0);

        //Result
        public Bitmap Bitmap;
        public Int32 Width;
        public Int32 Height;
        public Vector4[] Pixels;
        public World World;

        public Boolean UseShadows = true;

        // TODO: declare Vacuum model being model for empty space
        public Model Vacuum = new Model() { Shader = new Phong() };

        // TODO: declare MinFactor (=0.01) being the minimal color factor we trace rays. Rays with factor less than MinFactor are not traced
        public Double MinFactor = 0.01;

        // TODO: declare ReflectionCount and RefractionCount being the maximal number of reflection and refraction iterations.
        public Int32 ReflectionCount = 0;
        public Int32 RefractionCount = 0;

        // TODO: declare UseFresnel to control usage of Fresnel term evaluation 
        public Boolean UseFresnel = true;
        // TODO: declare FresnelPower (=1) being the power coefficient in the Schlick approximation of Fresnel term
        public Double FresnelPower = 1;


        // Native size of Bitmap (BitmapWidth, BitmapHeight)
        public int BitmapWidth, BitmapHeight;

        // SubSampleCount (=1) being the number of sub sampling rays per pixel
        public int SubSampleCount = 1;

        //Blur
        public Boolean UseBlur = false;
        public Int32 BlurRadius = 1;

        //DOF
        public bool UseDof;
        public double DofFactor = 1.0;
        public double DofZ = 13.0;


		#endregion


		#region Init

        public Camera(Int32 bitmapWidth, Int32 bitmapHeight)
		{
            BitmapWidth = bitmapWidth;
            BitmapHeight = bitmapHeight;
            Width = BitmapWidth * SubSampleCount;
            Height = BitmapHeight * SubSampleCount;
            Bitmap = new Bitmap(BitmapWidth, BitmapHeight);
            Pixels = new Vector4[Width * Height];
		}

		#endregion


		#region Buffer Acess

		public Vector4 GetPixel(Int32 i, Int32 j)
		{
			return Pixels[i + j * Width];
		}

		public void SetPixel(Int32 i, Int32 j, Vector4 color)
		{
			Pixels[i + j * Width] = color;
		}

		#endregion


		#region Rendering

		public void Render()
		{
            InitRender();
			RayTrace();
            AA();
			PresentFrame();
		}

        public void InitRender()
        {
            // Set total number of pixels (Width, Height) in Pixels buffer based on the SubSampleCount and size of bitmap
            Width = BitmapWidth * SubSampleCount;
            Height = BitmapHeight * SubSampleCount;

            Bitmap = new Bitmap(BitmapWidth, BitmapHeight);
            Pixels = new Vector4[Width * Height];
        }

        /// <summary>Derived from Computer Graphics - David Mount.
        /// </summary>
		public void RayTrace()
		{
            W = -1 * (Target - Position).Normalized;
            U = (Up % W).Normalized;
            V = (W % U);

            Double AspectRatio = (Double)Width / (Double)Height;
            Double h = 2 * Math.Tan(MathEx.DegToRad(FovY / 2.0));
            Double w = h * AspectRatio;

            for (int r = 0; r < Height; r++)
            {
                for (int c = 0; c < Width; c++)
                {
                    Double ur = h * ((Double)r / (Double)Height - 0.5);
                    Double uc = w * ((Double)c / (Double)Width - 0.5);
                    Vector4 rayDir = Position + (uc * U + ur * V - W);
                    rayDir = (rayDir - Position).Normalized;
                    Ray ray = new Ray(Position, rayDir, zNear, zFar);
                    Vector4 color = RayTrace(ray);
                    SetPixel(c, Height - 1 - r, color);
                }
            }
		}

        /// <summary>
        /// Ray trace the generated ray, compute the lighting, shadows, reflections and refractions
        /// </summary>
        /// <param name="ray">Traced ray</param>
        public Vector4 RayTrace(Ray ray)
        {
            Vector4 color = Vector4.Zero;
            World.Collide(ray);
            if (ray.HitModel == null) return BgColor;

            Vector4 hitPoint = ray.GetHitPoint();

            Vector4 hitNormal = new Vector4();
            //ToDo: Check if the normal has + or - direction
            //Vector4 hitNormal = ray.HitModel.Shader.GetNormal(hitPoint, ray.HitNormal);
            //hitNormal = (ray.HitNormal * hitNormal > 0) ? hitNormal : -hitNormal; 

            Ray lightRay = new Ray();
            Vector4 viewDir = -ray.Direction;

            foreach (Light light in World.Lights)
            {
                light.SetLightRayAt(hitPoint, lightRay);

                Double nl = hitNormal * lightRay.Direction;

                if (UseShadows && nl > 0.0f)
                    World.Collide(lightRay);

                Double intensityFactor = (lightRay.HitModel != null) ? 0.5 : 1;

                color += ray.HitModel.Shader.GetColor(hitPoint, hitNormal, viewDir, lightRay.Direction, intensityFactor * light.GetIntensityAt(hitPoint), light);
            }

            return color;
        }

        public void AA()
        {
            for (Int32 y = 0; y < BitmapHeight; y++)
            {
                for (Int32 x = 0; x < BitmapWidth; x++)
                {
                    Vector4 c = Vector4.Zero;
                    // calculated sub-sampled color by averaging colors of sub-pixels
                    for (Int32 jSub = 0; jSub < SubSampleCount; jSub++)
                    {
                        for (Int32 iSub = 0; iSub < SubSampleCount; iSub++)
                        {
                            c += GetPixel(x * SubSampleCount + iSub, y * SubSampleCount + jSub);
                        }
                    }
                    c *= (1.0 / (double)(SubSampleCount * SubSampleCount));
                    SetPixel(x, y, c);
                }
            }
        }

        /// <summary>
        /// Copy all the pixels from pixel buffer to the Bitmap.
        /// Color is clamped in post process.
        /// </summary>
        public void PresentFrame()
        {
            for (Int32 y = 0; y < BitmapHeight; y++)
            {
                for (Int32 x = 0; x < BitmapWidth; x++)
                {
                    Vector4 c = GetPixel(x, y);
                    c = Vector4.Clamp(c, 0, 1);
                    Bitmap.SetPixel(x, y, Color.FromArgb(
                        (Int32)(255 * c.X),
                        (Int32)(255 * c.Y),
                        (Int32)(255 * c.Z))
                    );
                }
            }
        }

		#endregion
	}
}
